// (ocean) water simulation and display (OpenGL)
// subsim (C)+(W) Thorsten Jordan. SEE LICENSE

// this file contains special x86-SSE optimized functions.
// nearly 40% of cpu time is consumed in ONE function,
// (compute_coord), so this is optimized with SSE.

#ifdef USE_SSE

#include "water_sse.h"

static const float fac_1_64_and_1_128[8] __attribute__((aligned(16))) = {
	1.0f/64, 1.0f/64, 1.0f/64, 1.0f/64,
	1.0f/128, 1.0f/128, 1.0f/128, 1.0f/128 };
static float v_sse[8] __attribute__((aligned(16)));
static float vadd_sse[8] __attribute__((aligned(16)));
static float vfrac_sse[8] __attribute__((aligned(16)));
static float vfracadd_sse[8] __attribute__((aligned(16)));
static float xyfrac2_sse[8] __attribute__((aligned(16)));
static const float one_sse[8] __attribute__((aligned(16))) = { 1.0f, 1.0f, 1.0f, 1.0f };
// const while water object exist:
static float rcpwavelength[4] __attribute__((aligned(16)));
static float waveresolution_sse[4] __attribute__((aligned(16)));
static const float vadd_fac01234444[8] __attribute__((aligned(16))) = { 0.0f, 1.0f, 2.0f, 3.0f,
									4.0f, 4.0f, 4.0f, 4.0f };
static const float fac2_16[4] __attribute__((aligned(16))) = {
	65536.0f, 65536.0f, 65536.0f, 65536.0f  };
static const Uint32 helpervalues[8] = { 0xfff, 0xfff, 0x3ff, 0x3ff, 2048, 2048, 512, 512 };
static Uint32 x0y0x1y1[16];
static Uint32* tmpptr[4];
static const Uint32 one_2x32[2] = { 0x1, 0x1 };
static Uint32 wave_res_and_2x[2];
static Uint32 waveresshiftq[2] = { 0, 0 };
static const Uint32 fac2_16i[2] = { 65536, 65536 };

// give four indices in mm0/mm1, fills xmm5/6/7 with heights/displx/disply
// be careful! it trashes mm0-7 !
static inline void get_height_and_displacement_sse(const Uint32* baseptr)
{
	// mm0/mm1 hold indices for four coordinates to load height/displx/disply from
	asm volatile ("pslld $2, %%mm0		\n\t"	// *4 to compute index of 32bit-words
		      "pslld $2, %%mm1		\n\t"	// *4 to compute index of 32bit-words
		      "movl %%eax, (%%esi)	\n\t"	// store base ptr address
		      "movl %%eax, 4(%%esi)	\n\t"	// store base ptr address, second dword
		      "movq (%%esi), %%mm2	\n\t"	// load base ptr address
		      "paddd %%mm2, %%mm0	\n\t"	// add address
		      "paddd %%mm2, %%mm1	\n\t"	// add address
		      "movq %%mm0, (%%esi)	\n\t"	// store addresses 0 and 1
		      "movq %%mm1, 8(%%esi)	\n\t"	// store addresses 2 and 3
		      : : "a" (baseptr), "S" (tmpptr)
		      : /* mm0-1 */ );

	asm volatile ("movl (%%esi), %%eax	\n\t"	// load address 0
		      "movl 4(%%esi), %%ebx	\n\t"	// load address 1
		      "movl 8(%%esi), %%ecx	\n\t"	// load address 2
		      "movl 12(%%esi), %%edx	\n\t"	// load address 3
		      "movl (%%eax), %%eax	\n\t"	// load value at address 0
		      "movl (%%ebx), %%ebx	\n\t"	// load value at address 1
		      "movl (%%ecx), %%ecx	\n\t"	// load value at address 2
		      "movl (%%edx), %%edx	\n\t"	// load value at address 3
		      "movl %%eax, (%%esi)	\n\t"	// store value of address 0
		      "movl %%ebx, 4(%%esi)	\n\t"	// store value of address 1
		      "movl %%ecx, 8(%%esi)	\n\t"	// store value of address 2
		      "movl %%edx, 12(%%esi)	\n\t"	// store value of address 3
		      "movq (%%esi), %%mm0	\n\t"	// load values of address 0 and 1
		      "movq 8(%%esi), %%mm1	\n\t"	// load values of address 2 and 3
		      "movq %%mm0, %%mm2	\n\t"	// copy for displx
		      "movq %%mm1, %%mm3	\n\t"	// copy for displx
		      "psrld $22, %%mm2		\n\t"	// >> 22 for displx
		      "psrld $22, %%mm3		\n\t"	// >> 22 for displx
		      "movq %%mm0, %%mm4	\n\t"	// copy for disply
		      "movq %%mm1, %%mm5	\n\t"	// copy for disply
		      "psrld $12, %%mm4		\n\t"	// >> 12 for disply
		      "psrld $12, %%mm5		\n\t"	// >> 12 for disply
		      "movq (%%edi), %%mm6	\n\t"	// load 0xfff
		      "movq 8(%%edi), %%mm7	\n\t"	// load 0x3ff
		      "pand %%mm6, %%mm0	\n\t"	// and with 0xfff
		      "pand %%mm6, %%mm1	\n\t"	// and with 0xfff
		      "pand %%mm7, %%mm2	\n\t"	// and with 0x3ff
		      "pand %%mm7, %%mm3	\n\t"	// and with 0x3ff
		      "pand %%mm7, %%mm4	\n\t"	// and with 0x3ff
		      "pand %%mm7, %%mm5	\n\t"	// and with 0x3ff
		      "movq 16(%%edi), %%mm6	\n\t"	// load 2048
		      "movq 24(%%edi), %%mm7	\n\t"	// load 512
		      "psubd %%mm6, %%mm0	\n\t"	// sub 2048
		      "psubd %%mm6, %%mm1	\n\t"	// sub 2048
		      "psubd %%mm7, %%mm2	\n\t"	// sub 512
		      "psubd %%mm7, %%mm3	\n\t"	// sub 512
		      "psubd %%mm7, %%mm4	\n\t"	// sub 512
		      "psubd %%mm7, %%mm5	\n\t"	// sub 512
		      // now mm0-5 hold values for height/displx/disply
		      "cvtpi2ps %%mm1, %%xmm5	\n\t"	// to float
		      "cvtpi2ps %%mm3, %%xmm6	\n\t"	// to float
		      "cvtpi2ps %%mm5, %%xmm7	\n\t"	// to float
		      "movlhps %%xmm5, %%xmm5	\n\t"	// and in upper 2 floats
		      "movlhps %%xmm6, %%xmm6	\n\t"	// and in upper 2 floats
		      "movlhps %%xmm7, %%xmm7	\n\t"	// and in upper 2 floats
		      "cvtpi2ps %%mm0, %%xmm5	\n\t"	// to float
		      "cvtpi2ps %%mm2, %%xmm6	\n\t"	// to float
		      "cvtpi2ps %%mm4, %%xmm7	\n\t"	// to float
		      : : "S" (tmpptr), "D" (helpervalues)
		      : "eax", "ebx", "ecx", "edx" /* and mm0-7*/ );

	// now multiply with factors
	asm volatile ("mulps 16(%%eax), %%xmm5	\n\t"	// multiply with 16/2048=1/128
		      "mulps (%%eax), %%xmm6	\n\t"	// multiply with 8/512=1/64
		      "mulps (%%eax), %%xmm7	\n\t"	// multiply with 8/512=1/64
		      : : "a" (fac_1_64_and_1_128) : );
}



bool x86_sse_supported()
{
	Uint32 cpuid[4];
	asm volatile ("movl $1, %%eax		\n\t"
		      "cpuid			\n\t"
		      "mov %%eax, (%%esi)	\n\t"
		      "mov %%ebx, 4(%%esi)	\n\t"
		      "mov %%ecx, 8(%%esi)	\n\t"
		      "mov %%edx, 12(%%esi)	\n\t"
		      : : "S" (&cpuid) : "eax", "ebx", "ecx", "edx" );
	// bit# 23: mmx
	// bit# 25: sse
	// bit# 26: sse2
	return ((cpuid[3] >> 25) & 1) != 0;
}



void water_compute_coord_init_sse(float wavetile_length_rcp, unsigned wave_resolution,
				  unsigned wave_resolution_shift)
{
	rcpwavelength[0] = rcpwavelength[1] = rcpwavelength[2] = rcpwavelength[3] =
		wavetile_length_rcp;
	waveresolution_sse[0] = waveresolution_sse[1] = waveresolution_sse[2] = waveresolution_sse[3] = 
		float(wave_resolution);
	wave_res_and_2x[0] = wave_res_and_2x[1] = wave_resolution-1;
	waveresshiftq[0] = wave_resolution_shift;
}



void water_compute_coord_1line_sse(const vector2f& v, const vector2f& vadd,
				   const vector2f& vfrac, const vector2f& vfracadd,
				   const Uint32* tiledata_baseptr, float minusviewposz,
				   float* coordptr, unsigned xres)
{
	// preparation part 1: fill v.x/y, vadd.x/y
	asm volatile ("movss (%%eax), %%xmm0		\n\t"	// load v.x
		      "movss 4(%%eax), %%xmm1		\n\t"	// load v.y
		      "shufps $0x00, %%xmm0, %%xmm0	\n\t"	// set v.x to all 4 floats
		      "shufps $0x00, %%xmm1, %%xmm1	\n\t"	// set v.y to all 4 floats
		      "movss (%%ebx), %%xmm2		\n\t"	// load vadd.x
		      "movss 4(%%ebx), %%xmm3		\n\t"	// load vadd.y
		      "shufps $0x00, %%xmm2, %%xmm2	\n\t"	// set vadd.x to all 4 floats
		      "shufps $0x00, %%xmm3, %%xmm3	\n\t"	// set vadd.y to all 4 floats
		      "movaps 16(%%edx), %%xmm4		\n\t"	// load 4
		      "movaps %%xmm4, %%xmm5		\n\t"	// load 4
		      "mulps %%xmm2, %%xmm4		\n\t"	// vadd.x*4
		      "mulps %%xmm3, %%xmm5		\n\t"	// vadd.x*4
		      "movaps %%xmm4, (%%ecx)		\n\t"	// store vadd.x*4 ps
		      "movaps %%xmm5, 16(%%ecx)		\n\t"	// store vadd.y*4 ps
		      "mulps (%%edx), %%xmm2		\n\t"	// mul vadd.x with 0/1/2/3
		      "mulps (%%edx), %%xmm3		\n\t"	// mul vadd.y with 0/1/2/3
		      "addps %%xmm2, %%xmm0		\n\t"	// add to v.x to get start vector
		      "addps %%xmm3, %%xmm1		\n\t"	// add to v.y to get start vector
		      "movaps %%xmm0, (%%esi)		\n\t"	// store v_sse
		      "movaps %%xmm1, 16(%%esi)		\n\t"	// store v_sse
		      : /* output */
		      : /* input */ "a" (&v.x), "b" (&vadd.x), "c" (vadd_sse),
			"d" (vadd_fac01234444), "S" (v_sse)
		      : /* manipulated */ );

	// preparation part 2: fill vfrac.x/y, vfracadd.x/y
	asm volatile ("movss (%%eax), %%xmm0		\n\t"	// load vfrac.x
		      "movss 4(%%eax), %%xmm1		\n\t"	// load vfrac.y
		      "shufps $0x00, %%xmm0, %%xmm0	\n\t"	// set vfrac.x to all 4 floats
		      "shufps $0x00, %%xmm1, %%xmm1	\n\t"	// set vfrac.y to all 4 floats
		      "movss (%%ebx), %%xmm2		\n\t"	// load vfracadd.x
		      "movss 4(%%ebx), %%xmm3		\n\t"	// load vfracadd.y
		      "shufps $0x00, %%xmm2, %%xmm2	\n\t"	// set vfracadd.x to all 4 floats
		      "shufps $0x00, %%xmm3, %%xmm3	\n\t"	// set vfracadd.y to all 4 floats
		      "movaps 16(%%edx), %%xmm4		\n\t"	// load 4
		      "movaps %%xmm4, %%xmm5		\n\t"	// load 4
		      "mulps %%xmm2, %%xmm4		\n\t"	// vfracadd.x*4
		      "mulps %%xmm3, %%xmm5		\n\t"	// vfracadd.x*4
		      "movaps %%xmm4, (%%ecx)		\n\t"	// store vfracadd.x*4 ps
		      "movaps %%xmm5, 16(%%ecx)		\n\t"	// store vfracadd.y*4 ps
		      "mulps (%%edx), %%xmm2		\n\t"	// mul vfracadd.x with 0/1/2/3
		      "mulps (%%edx), %%xmm3		\n\t"	// mul vfracadd.y with 0/1/2/3
		      "addps %%xmm2, %%xmm0		\n\t"	// add to vfrac.x to get start vector
		      "addps %%xmm3, %%xmm1		\n\t"	// add to vfrac.y to get start vector
		      "movaps %%xmm0, (%%esi)		\n\t"	// store vfrac_sse
		      "movaps %%xmm1, 16(%%esi)		\n\t"	// store vfrac_sse
		      : /* output */
		      : /* input */ "a" (&vfrac.x), "b" (&vfracadd.x), "c" (vfracadd_sse),
			"d" (vadd_fac01234444), "S" (vfrac_sse)
		      : /* manipulated */ );

/*
  sometimes SSE and C++ generated values differ by a large amount (~ 0.5... 1 total!)
  SSE is much less precise because all operands are stored with only float precision
  leading to serious rounding errors. But can it be THAT unprecise? fixme, check!
  yes, maybe: small rounding error in x/y direction can lead to large jumps in per tile x/y values
  for far coordinates. With larger variance of x/y coordinates different height values can be fetched for
  far coordinates, leading to greater differences in height values.
*/

	// note! works only if xres+1 is multiple of 4! that means xres % 4 = 3
	unsigned xloopc = (xres+1)/4;
	for (unsigned xx = 0; xx < xloopc; ++xx, coordptr += 12) {

		// compute xfrac2/yfrac2 to xmm0/xmm1
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load vfrac.x
			      "movaps 16(%%eax), %%xmm1		\n\t"	// load vfrac.y
			      "cvttps2pi %%xmm0, %%mm4		\n\t"	// trunc to integer in mm4
			      "cvttps2pi %%xmm1, %%mm6		\n\t"	// trunc to integer in mm6
			      "movhlps %%xmm0, %%xmm2		\n\t"	// move 2 upper floats to 2 lower in xmm2
			      "movhlps %%xmm1, %%xmm3		\n\t"	// move 2 upper floats to 2 lower in xmm3
			      "cvttps2pi %%xmm2, %%mm5		\n\t"	// trunc to integer in mm5
			      "cvttps2pi %%xmm3, %%mm7		\n\t"	// trunc to integer in mm7
			      "movq (%%ecx), %%mm0		\n\t"	// load wave_resolution-1
			      "movq %%mm0, %%mm1		\n\t"	// load wave_resolution-1
			      "movq %%mm0, %%mm2		\n\t"	// load wave_resolution-1
			      "movq %%mm0, %%mm3		\n\t"	// load wave_resolution-1
			      "pand %%mm4, %%mm0		\n\t"	// mask out tile coordinates
			      "pand %%mm5, %%mm1		\n\t"	// mask out tile coordinates
			      "pand %%mm6, %%mm2		\n\t"	// mask out tile coordinates
			      "pand %%mm7, %%mm3		\n\t"	// mask out tile coordinates
			      "psubd %%mm0, %%mm4		\n\t"	// subtract masked out part
			      "psubd %%mm1, %%mm5		\n\t"	// " | mm4-7 will then hold the part
			      "psubd %%mm2, %%mm6		\n\t"	// " | of coordinates that are greater
			      "psubd %%mm3, %%mm7		\n\t"	// " | than one tile (when vfrac+vfracadd > 1tile)
			      "cvtpi2ps %%mm5, %%xmm2		\n\t"	// back to float
			      "cvtpi2ps %%mm7, %%xmm3		\n\t"	// back to float
			      "movlhps %%xmm2, %%xmm2		\n\t"	// and in upper 2 floats
			      "movlhps %%xmm3, %%xmm3		\n\t"	// and in upper 2 floats
			      "cvtpi2ps %%mm4, %%xmm2		\n\t"	// back to float
			      "cvtpi2ps %%mm6, %%xmm3		\n\t"	// back to float
			      "subps %%xmm2, %%xmm0		\n\t"	// subtract >1tile part from vfrac
			      "subps %%xmm3, %%xmm1		\n\t"	// subtract >1tile part from vfrac
			      "movaps %%xmm0, (%%eax)		\n\t"	// store corrected vfrac.x
			      "movaps %%xmm1, 16(%%eax)		\n\t"	// store corrected vfrac.y
			      "cvtpi2ps %%mm1, %%xmm2		\n\t"	// back to float for x0/y0
			      "cvtpi2ps %%mm3, %%xmm3		\n\t"	// back to float
			      "movlhps %%xmm2, %%xmm2		\n\t"	// and in upper 2 floats
			      "movlhps %%xmm3, %%xmm3		\n\t"	// and in upper 2 floats
			      "cvtpi2ps %%mm0, %%xmm2		\n\t"	// back to float
			      "cvtpi2ps %%mm2, %%xmm3		\n\t"	// back to float
			      "subps %%xmm2, %%xmm0		\n\t"	// now xmm0 holds its fractional part
			      "subps %%xmm3, %%xmm1		\n\t"	// now xmm1 holds its fractional part
			      // now mm0/1 holds x0, mm2/3 holds y0

			      // compute x1/y1 in mm4-7
			      "movq (%%ebx), %%mm4		\n\t"	// load 1
			      "movq %%mm4, %%mm5		\n\t"	// load 1
			      "movq %%mm4, %%mm6		\n\t"	// load 1
			      "movq %%mm4, %%mm7		\n\t"	// load 1
			      "paddd %%mm0, %%mm4		\n\t"	// x1=1+x0
			      "paddd %%mm1, %%mm5		\n\t"	// x1=1+x0
			      "paddd %%mm2, %%mm6		\n\t"	// y1=1+y0
			      "paddd %%mm3, %%mm7		\n\t"	// y1=1+y0
			      "pand (%%ecx), %%mm4 		\n\t"	// and width-1
			      "pand (%%ecx), %%mm5 		\n\t"	// and width-1
			      "pand (%%ecx), %%mm6 		\n\t"	// and width-1
			      "pand (%%ecx), %%mm7 		\n\t"	// and width-1

			      // shift y0/y1 by wave_resolution_shift
			      "pslld (%%edx), %%mm2		\n\t"	// shift y0
			      "pslld (%%edx), %%mm3		\n\t"	// shift y0
			      "pslld (%%edx), %%mm6		\n\t"	// shift y1
			      "pslld (%%edx), %%mm7		\n\t"	// shift y1

			      // save x0/y0/x1/y1
			      "movq %%mm0, (%%esi)		\n\t"	// save x0
			      "movq %%mm1, 8(%%esi)		\n\t"	// save x0
			      "movq %%mm2, 16(%%esi)		\n\t"	// save y0
			      "movq %%mm3, 24(%%esi)		\n\t"	// save y0
			      "movq %%mm4, 32(%%esi)		\n\t"	// save x1
			      "movq %%mm5, 40(%%esi)		\n\t"	// save x1
			      "movq %%mm6, 48(%%esi)		\n\t"	// save y1
			      "movq %%mm7, 56(%%esi)		\n\t"	// save y1
			      : : "a" (vfrac_sse), "b" (one_2x32), "c" (wave_res_and_2x),
				"d" (waveresshiftq), "S" (x0y0x1y1) : );

		// save xfrac2/yfrac2
		asm volatile ("movaps %%xmm0, (%%eax)		\n\t" // store xfrac2
			      "movaps %%xmm1, 16(%%eax)		\n\t" // store yfrac2
			      "movaps (%%ebx), %%xmm0		\n\t" // load 1.0
			      "movaps (%%ebx), %%xmm1		\n\t" // load 1.0
			      "subps (%%eax), %%xmm0		\n\t" // 1.0-xfrac2
			      "subps 16(%%eax), %%xmm1		\n\t" // 1.0-yfrac2
			      : : "a" (xyfrac2_sse), "b" (one_sse) : );

		// compute i0 etc. xmm0/1 holds xfrac2/yfrac2
		asm volatile ("paddd %%mm2, %%mm0		\n\t"	// x0+y0
			      "paddd %%mm3, %%mm1		\n\t"	// x0+y0
			      : : : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("mulps %%xmm1, %%xmm0		\n\t"	// comp (1-xfrac2)*(1-yfrac2)
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "movaps %%xmm5, %%xmm2		\n\t"
			      "movaps %%xmm6, %%xmm3		\n\t"
			      "movaps %%xmm7, %%xmm4		\n\t"
			      : : : );

		// compute i1 etc.
		asm volatile ("movq 32(%%eax), %%mm0		\n\t"	// get x1
			      "movq 40(%%eax), %%mm1		\n\t"	// get x1
			      "paddd 16(%%eax), %%mm0		\n\t"	// add y0
			      "paddd 24(%%eax), %%mm1		\n\t"	// add y0
			      : : "a" (x0y0x1y1) : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load xfrac2
			      "mulps %%xmm1, %%xmm0		\n\t"	// comp xfrac2*(1-yfrac2)
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "addps %%xmm5, %%xmm2		\n\t"
			      "addps %%xmm6, %%xmm3		\n\t"
			      "addps %%xmm7, %%xmm4		\n\t"
			      : : "a" (xyfrac2_sse) : );

		// compute i2 etc.
		asm volatile ("movq (%%eax), %%mm0			\n\t"	// get x0
			      "movq 8(%%eax), %%mm1			\n\t"	// get x0
			      "paddd 48(%%eax), %%mm0		\n\t"	// add y1
			      "paddd 56(%%eax), %%mm1		\n\t"	// add y1
			      : : "a" (x0y0x1y1) : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("movaps (%%ebx), %%xmm0		\n\t"	// load 1
			      "movaps 16(%%eax), %%xmm1		\n\t"	// load yfrac2
			      "subps (%%eax), %%xmm0		\n\t"	// comp 1-xfrac2
			      "mulps %%xmm1, %%xmm0		\n\t"	// comp (1-xfrac2)*yfrac
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "addps %%xmm5, %%xmm2		\n\t"
			      "addps %%xmm6, %%xmm3		\n\t"
			      "addps %%xmm7, %%xmm4		\n\t"
			      : : "a" (xyfrac2_sse), "b" (one_sse) : );

		// compute i3 etc.
		asm volatile ("movq 32(%%eax), %%mm0		\n\t"	// get x1
			      "movq 40(%%eax), %%mm1		\n\t"	// get x1
			      "paddd 48(%%eax), %%mm0		\n\t"	// add y1
			      "paddd 56(%%eax), %%mm1		\n\t"	// add y1
			      : : "a" (x0y0x1y1) : );
		get_height_and_displacement_sse(tiledata_baseptr);
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load xfrac2
			      "mulps %%xmm1, %%xmm0		\n\t"	// comp xfrac2*yfrac2
			      "mulps %%xmm0, %%xmm5		\n\t"
			      "mulps %%xmm0, %%xmm6		\n\t"
			      "mulps %%xmm0, %%xmm7		\n\t"
			      "addps %%xmm5, %%xmm2		\n\t"
			      "addps %%xmm6, %%xmm3		\n\t"
			      "addps %%xmm7, %%xmm4		\n\t"
			      : : "a" (xyfrac2_sse) : );

		// now add xyzpos, that means add vx/vy to xmm3/4 and -viewpos.z to xmm2
		asm volatile ("movss (%%eax), %%xmm5		\n\t"	// load -viewpos.z
			      "shufps $0x00, %%xmm5, %%xmm5	\n\t"	// load to all ps
			      "addps (%%ebx), %%xmm3		\n\t"	// add v.x
			      "addps 16(%%ebx), %%xmm4		\n\t"	// add v.y
			      "addps %%xmm5, %%xmm2		\n\t"	// add -viewpos.z
			      : : "a" (&minusviewposz), "b" (v_sse) : );

		// now xmm2-4 holds interpolated height/displx/disply
		// coordptr points to first of 3*4 floats to store results to.
		// xmm2-4 hold values    : z3 z2 z1 z0 | x3 x2 x1 x0 | y3 y2 y1 y0
		// stored should be      : x1 z0 y0 x0 | y2 x2 z1 y1 | z3 y3 x3 z2
		// intermediate form     : x1 x0 z0 y0 | y2 y1 z1 x2 | y3 x3 z3 z2
		// mix registers in xmm5-7 and store result
		asm volatile ("movaps %%xmm4, %%xmm5		\n\t" //5: y3y2y1y0, copy
			      "movaps %%xmm2, %%xmm6		\n\t" //6: z3z2z1z0, copy
			      "unpcklps %%xmm2, %%xmm5		\n\t" //5: z1y1z0y0, mix z0y0
			      "shufps $0xe4, %%xmm3, %%xmm6	\n\t" //6: x3x2z1z0, mix z1x2 part1
			      "movlhps %%xmm3, %%xmm5		\n\t" //5: x1x0z0y0, copy x1x0
			      "shufps $0x66, %%xmm6, %%xmm6	\n\t" //6: z1x2z1x2, mix z1x2 part2
			      "movaps %%xmm3, %%xmm7		\n\t" //7: x3x2x1x0, copy
			      "shufps $0xd2, %%xmm5, %%xmm5	\n\t" //5: x1z0y0x0, xmm5 ready
			      "unpckhps %%xmm4, %%xmm7		\n\t" //7: y3x3y2x2, get y3x3
			      "shufps $0x94, %%xmm4, %%xmm6	\n\t" //6: y2y1z1x2, copy y2y1
			      "movhlps %%xmm2, %%xmm7		\n\t" //7: y3x3z3z2, copy z3z2
			      "shufps $0xc6, %%xmm6, %%xmm6	\n\t" //6: y2x2z1y1, xmm6 ready
			      "shufps $0x78, %%xmm7, %%xmm7	\n\t" //7: z3y3x3z2, xmm7 ready
			      "movups %%xmm5, (%%eax)		\n\t" // store bytes 0-3
			      "movups %%xmm6, 16(%%eax)		\n\t" // store bytes 4-7
			      "movups %%xmm7, 32(%%eax)		\n\t" // store bytes 8-11
			      : : "a" (coordptr) : );

		// increase v/vfrac
		asm volatile ("movaps (%%eax), %%xmm0		\n\t"	// load vadd.x
			      "movaps 16(%%eax), %%xmm1		\n\t"	// load vadd.y
			      "movaps (%%ecx), %%xmm2		\n\t"	// load vfracadd.x
			      "movaps 16(%%ecx), %%xmm3		\n\t"	// load vfracadd.y
			      "addps (%%ebx), %%xmm0		\n\t"	// add to v.x
			      "addps 16(%%ebx), %%xmm1		\n\t"	// add to v.y
			      "addps (%%edx), %%xmm2		\n\t"	// add to vfrac.x
			      "addps 16(%%edx), %%xmm3		\n\t"	// add to vfrac.y
			      "movaps %%xmm0, (%%ebx)		\n\t"	// store new v.x
			      "movaps %%xmm1, 16(%%ebx)		\n\t"	// store new v.y
			      "movaps %%xmm2, (%%edx)		\n\t"	// store new vfrac.x
			      "movaps %%xmm3, 16(%%edx)		\n\t"	// store new vfrac.y
			      : /* output */
			      : /* input */ "a" (vadd_sse), "b" (v_sse), "c" (vfracadd_sse), "d" (vfrac_sse)
			      : /* manipulated */ );
	}

	asm volatile ("emms		\n\t");
}

#endif
